//========= Copyright  2010, fire64 LLC, All rights reserved. ============
//
// Purpose: 
//
// $NoKeywords: $
//=============================================================================
#include "stdafx.h"
#include "Csocket.h"

//-----------------------------------------------------------------------------
// Purpose: Constructor
//-----------------------------------------------------------------------------
Csocket::Csocket( ESocketProtocol protocol)
{
	if( protocol == eSocketProtocolTCP )
	{
		m_iftcp = true;

		current_sock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP );

	}
	else if( protocol == eSocketProtocolUDP )
	{
		m_iftcp = false;

		current_sock = socket ( AF_INET, SOCK_DGRAM, IPPROTO_UDP ); ;
	}
	else if( protocol == eSocketProtocolIP )
	{
		m_iftcp = true;

		current_sock = socket ( AF_INET, SOCK_STREAM, IPPROTO_IP );
	}

	m_usetimeout = false;
}

//-----------------------------------------------------------------------------
// Purpose: Destructor
//-----------------------------------------------------------------------------
Csocket::~Csocket()
{
	closesocket( current_sock );
}

//-----------------------------------------------------------------------------
// Purpose: binding port
//-----------------------------------------------------------------------------
int Csocket::BindPort( int port )
{
	int res;

    dest_addr.sin_family=AF_INET;
    dest_addr.sin_port=htons ( port );
	dest_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	res = bind (current_sock, (sockaddr *)&dest_addr, sizeof(dest_addr));

	return res;
}

//-----------------------------------------------------------------------------
// Purpose: get ip by host
//-----------------------------------------------------------------------------
unsigned long Csocket::resolve(char *host)
{
    unsigned long ret;
    struct hostent * hp = gethostbyname(host);
    if (!hp) ret = inet_addr(host);
    if ((!hp)&&(ret == INADDR_NONE)) return 0;
    if (hp != NULL) memcpy((void*)&ret, hp->h_addr,hp->h_length);
    return ret;
}

//-----------------------------------------------------------------------------
// Purpose: set adr for socket
//-----------------------------------------------------------------------------
int Csocket::SetAdr( sockaddr_in  new_addr )
{
	dest_addr = new_addr;

	return 1;
}

//-----------------------------------------------------------------------------
// Purpose: set adr for socket
//-----------------------------------------------------------------------------
int Csocket::SetAdr( char *ip, int port )
{
	dest_addr.sin_family=AF_INET;
    dest_addr.sin_port=htons ( port );
	dest_addr.sin_addr.s_addr = resolve(ip);

	return 1;
}

//-----------------------------------------------------------------------------
// Purpose: set adr for socket
//-----------------------------------------------------------------------------
bool Csocket::Connect( )
{
	if( IfUseTimeOut() )
	{
		// Estabilish a TCP connection
		connect(current_sock, (sockaddr*)&dest_addr, sizeof(dest_addr));

		fd_write.fd_array[0] = current_sock; fd_write.fd_count = 1;
		fd_error.fd_array[0] = current_sock; fd_error.fd_count = 1;
		if (select(0, 0, &fd_write, &fd_error, &timeout) > 0) 
		{
			if (fd_write.fd_count > 0)
			{
				return true;
			}
		}
		
		return false;
	}
	else
	{
		if( connect(current_sock, (sockaddr*)&dest_addr, sizeof(dest_addr)) == -1 )
		{
			return false;
		}
		else
		{
			return true;
		}
	}
}

//-----------------------------------------------------------------------------
// Purpose: Send Packet
//-----------------------------------------------------------------------------
bool Csocket::Send( char *buff, int sizebuff )
{
	int bufflen;
	
	if( !sizebuff )
	{
		bufflen = strlen(buff);
	}
	else
	{
		bufflen = sizebuff;
	}
	
	if( GetCurentProtocol( ) == eSocketProtocolTCP )
	{
		if (send(current_sock, buff, bufflen, 0) == SOCKET_ERROR)
		{
			return false;
		}
		else
		{
			return true;
		}
	}
	else
	{
		if (sendto(current_sock, buff, bufflen, 0, (sockaddr*)&dest_addr, sizeof(dest_addr) ) == SOCKET_ERROR)
		{
			return false;
		}
		else
		{
			return true;
		}
	}
}

//-----------------------------------------------------------------------------
// Purpose: Recv Packet
//-----------------------------------------------------------------------------
int Csocket::Recv( char *recvbuff, int size)
{
	int recvbytes;
	if (size==0)
		size = MAXRECVSIZE;

	if( IfUseTimeOut() )
	{
		FD_ZERO( &fd_read );
		FD_SET( current_sock, &fd_read );

		if( GetCurentProtocol( ) == eSocketProtocolTCP )
		{
			if( select(current_sock +1, &fd_read, NULL, NULL, &timeout) < 1 )
			{
				return 0;
			}
			else
			{
				recvbytes = recv(current_sock, recvbuff, size, 0);
			}
		}
		else
		{
			int server_addr_size = sizeof(dest_addr);

			if( select(current_sock +1, &fd_read, NULL, NULL, &timeout) < 1 )
			{
				return 0;
			}
			else
			{
				recvbytes = recvfrom(current_sock, recvbuff, size, 0, (sockaddr*) &dest_addr, &server_addr_size);
			}
		}
	}
	else
	{	
		if( GetCurentProtocol( ) == eSocketProtocolTCP )
		{
			recvbytes = recv(current_sock, recvbuff, size, 0);
		}
		else
		{
			int server_addr_size = sizeof(dest_addr);

			recvbytes = recvfrom(current_sock, recvbuff, size, 0, (sockaddr*) &dest_addr, &server_addr_size);
		}
	}

	return recvbytes;
}

//-----------------------------------------------------------------------------
// Purpose: Send UINT32 and convert to BigEndian
//-----------------------------------------------------------------------------
bool Csocket::SendInt32(UINT32 buff, bool reverse)
{
	if (reverse)
		buff = htonl(buff);
	return Send((char*)&buff, 4);
}

//-----------------------------------------------------------------------------
// Purpose: Send UINT16 and convert to BigEndian
//-----------------------------------------------------------------------------
bool Csocket::SendInt16(UINT16 buff, bool reverse)
{
	if (reverse)
		buff = htons(buff);
	return Send((char*)&buff, 2);
}

//-----------------------------------------------------------------------------
// Purpose: Recv UINT32 and convert to BigEndian
//-----------------------------------------------------------------------------
bool Csocket::RecvInt32(UINT32 *buff, bool reverse)
{
	if (Recv((char*)buff, 4) != 4)
		return false;
	if (reverse)
		*buff = htonl(*buff);
	return true;
}

//-----------------------------------------------------------------------------
// Purpose: Recv UINT16 and convert to BigEndian
//-----------------------------------------------------------------------------
bool Csocket::RecvInt16(UINT16 *buff, bool reverse)
{
	if (Recv((char*)buff, 2) != 4)
		return false;
	if (reverse)
		*buff = htons(*buff);
	return true;
}

//-----------------------------------------------------------------------------
// Purpose: TCP Timeout
//-----------------------------------------------------------------------------
int Csocket::SetTimeOut( int settimeout )
{
	timeout.tv_sec = settimeout / 1000;
	timeout.tv_usec = settimeout % 1000;
	u_long nNonBlocking = 1;

	m_usetimeout = true;

	// Set socket to non-blocking mode
	ioctlsocket(current_sock, FIONBIO, &nNonBlocking);

	return 0;
}

//-----------------------------------------------------------------------------
// Purpose: If use timeout
//-----------------------------------------------------------------------------
bool Csocket::IfUseTimeOut( )
{
	return m_usetimeout;
}

//-----------------------------------------------------------------------------
// Purpose: Accept TCP Connect
//-----------------------------------------------------------------------------
Csocket *Csocket::AcceptConnect( )
{
	int server_addr_size = sizeof(dest_addr);

	if( IfUseTimeOut() )
	{
		FD_ZERO( &fd_read );
		FD_SET( current_sock, &fd_read );

		if( GetCurentProtocol( ) == eSocketProtocolTCP )
		{
			select(current_sock +1, &fd_read, &fd_read, NULL, &timeout);
		}
	}
	
	int Sock = accept( current_sock, (sockaddr*) &dest_addr, &server_addr_size );
	if (Sock == SOCKET_ERROR)
	{
		return NULL;
	}
	else
	{
		Csocket *res = new Csocket(eSocketProtocolTCP);
		res->SetAdr(dest_addr);
		res->current_sock = Sock;
		return res;
	}
}


//-----------------------------------------------------------------------------
// Purpose: Listen TCP Server
//-----------------------------------------------------------------------------
bool Csocket::ListenServer( )
{
    if(listen(current_sock,10)!=0)
    {
        return false;
    }

	return true;
}

ESocketProtocol Csocket::GetCurentProtocol( )
{
	ESocketProtocol protocol;

	if( m_iftcp )
	{
		protocol = eSocketProtocolTCP;
	}
	else
	{
		protocol = eSocketProtocolUDP;
	}

	return protocol;
}

//-----------------------------------------------------------------------------
// Purpose: returned IP and Port of connected client (on format "xxx.xxx.xxx.xxx:xxxxx")
//-----------------------------------------------------------------------------
char *Csocket::GetAddr()
{
	sockaddr_in Addr;
	int len = sizeof(sockaddr_in);
	if (getsockname(current_sock, (sockaddr*)&Addr, &len) == SOCKET_ERROR)
	{
		return "0.0.0.0:0";
	}
	else
	{
		//return inet_ntoa(Addr.sin_addr);
		char *buf = new char[23];
		sprintf_s(buf, 23, "%s:%i", inet_ntoa(Addr.sin_addr), htons(Addr.sin_port));
		return buf;
	}
}

UINT32 Csocket::GetIP()
{
	sockaddr_in Addr;
	int len = sizeof(sockaddr_in);
	if (getsockname(current_sock, (sockaddr*)&Addr, &len) == SOCKET_ERROR)
	{
		return 0;
	}
	else
	{
		return Addr.sin_addr.s_addr;
	}
}

char *Csocket::IP2Str(UINT32 IP)
{
	in_addr addr;
	addr.s_addr = IP;
	return inet_ntoa(addr);
}

void Csocket::init()
{
	WSADATA wsaData;
	WSAStartup(MAKEWORD(2,2), &wsaData);
}

void Csocket::deinit()
{
	WSACleanup();
}